"""The Notion API mapped to Python objects by Pydantic.

The code was taken originally from [Notional] by Jason Heddings and is MIT-licensed.
Due to the Pydantic v2 migration and several other design changes, it was refactored since then.

[Notional]: https://github.com/jheddings/notional
"""
